import json
import re

import pytest

from pika import BasicProperties, DeliveryMode
from pika.spec import Basic

from hive.messaging import Channel, Message


class MockPika:
    def __getattr__(self, attr):
        if attr == "_prefetch_count":
            raise AttributeError(attr)
        raise NotImplementedError(attr)


class MockMethod:
    def __init__(self, *, returns=None):
        self.call_log = []
        self._returns = returns

    def __call__(self, *args, **kwargs):
        self.call_log.append((args, kwargs))
        return self._returns


class MockCallbackV1(MockMethod):
    """Pika-style on-message callback.
    """
    _warning_filter = "ignore:Pika-style .* callbacks:DeprecationWarning"

    def __call__(
            self,
            channel: Channel,
            method: Basic.Deliver,
            properties: BasicProperties,
            body: bytes,
    ):
        return super().__call__(channel, Message(method, properties, body))


class MockCallbackV2(MockMethod):
    """New-style on-message callback.
    """
    def __call__(
            self,
            channel: Channel,
            message: Message,
    ):
        return super().__call__(channel, message)


expect_properties = BasicProperties(
    content_type="application/json",
    delivery_mode=DeliveryMode.Persistent,
)


def test_publish_request():
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_publish = MockMethod()

    channel = Channel(pika=mock)
    channel.publish_request(
        message={
            "hello": "world",
        },
        routing_key="hallo.wereld",
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.requests",
        "exchange_type": "direct",
        "durable": True,
    })]
    assert mock.basic_publish.call_log == [((), {
        "exchange": "hive.requests",
        "routing_key": "hallo.wereld",
        "body": b'{"hello": "world"}',
        "properties": expect_properties,
        "mandatory": True,
    })]


@pytest.mark.parametrize("callback_cls", (MockCallbackV1, MockCallbackV2))
@pytest.mark.filterwarnings(MockCallbackV1._warning_filter)
def test_consume_requests(callback_cls):
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_qos = MockMethod()
    mock.queue_declare = MockMethod(
        returns=type("Result", (), dict(
            method=type("Method", (), dict(
                queue="TeStQuEu3")))))
    mock.queue_bind = MockMethod()
    mock.basic_consume = MockMethod()
    on_message_callback = callback_cls()
    mock.basic_ack = MockMethod()

    channel = Channel(pika=mock)
    channel.consume_requests(
        queue="arr.pirates",
        on_message_callback=on_message_callback,
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.requests",
        "exchange_type": "direct",
        "durable": True,
    }), ((), {
        "exchange": "hive.dead.letter",
        "exchange_type": "direct",
        "durable": True,
    })]
    assert mock.basic_qos.call_log == [((), {
        "prefetch_count": 1,
    })]
    assert mock.queue_declare.call_log == [((
        "x.arr.pirates",
    ), {
        "durable": True,
    }), ((
        "arr.pirates",
    ), {
        "durable": True,
        "arguments": {
            "x-dead-letter-exchange": "hive.dead.letter",
        },
    })]
    assert mock.queue_bind.call_log == [((), {
        "queue": "x.arr.pirates",
        "exchange": "hive.dead.letter",
        "routing_key": "arr.pirates",
    }), ((), {
        "queue": "arr.pirates",
        "exchange": "hive.requests",
        "routing_key": "arr.pirates",
    })]

    assert len(mock.basic_consume.call_log) == 1
    assert len(mock.basic_consume.call_log[0]) == 2
    got_callback = mock.basic_consume.call_log[0][1]["on_message_callback"]
    assert mock.basic_consume.call_log == [((), {
        "queue": "arr.pirates",
        "on_message_callback": got_callback,
    })]
    assert on_message_callback.call_log == []
    assert mock.basic_ack.call_log == []

    expect_method = type("method", (), {"delivery_tag": 5})
    expect_body = b'{"hello":"W0RLD"}'
    got_callback(channel._pika, expect_method, expect_properties, expect_body)

    assert len(on_message_callback.call_log) == 1
    assert len(on_message_callback.call_log[0]) == 2
    assert len(on_message_callback.call_log[0][0]) == 2
    message = on_message_callback.call_log[0][0][1]
    assert on_message_callback.call_log == [((channel, message), {})]
    assert message.method is expect_method
    assert message.properties is expect_properties
    assert message.body is expect_body

    assert mock.basic_ack.call_log == [((), {"delivery_tag": 5})]


def test_publish_mandatory_event():
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_publish = MockMethod()

    channel = Channel(pika=mock)
    channel.publish_event(
        message={
            "hello": "world",
        },
        routing_key="hoolloo.wooreled",
        mandatory=True,
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.events",
        "exchange_type": "direct",
        "durable": True,
    })]
    assert mock.basic_publish.call_log == [((), {
        "exchange": "hive.events",
        "routing_key": "hoolloo.wooreled",
        "body": b'{"hello": "world"}',
        "properties": expect_properties,
        "mandatory": True,
    })]


@pytest.mark.parametrize("callback_cls", (MockCallbackV1, MockCallbackV2))
@pytest.mark.filterwarnings(MockCallbackV1._warning_filter)
def test_consume_mandatory_events(callback_cls):
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_qos = MockMethod()
    mock.queue_declare = MockMethod(
        returns=type("Result", (), dict(
            method=type("Method", (), dict(
                queue="TeStQuEu3")))))
    mock.queue_bind = MockMethod()
    mock.basic_consume = MockMethod()
    on_message_callback = callback_cls()
    mock.basic_ack = MockMethod()

    channel = Channel(pika=mock)
    channel.consume_events(
        queue="arr.pirates",
        on_message_callback=on_message_callback,
        mandatory=True,
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.events",
        "exchange_type": "direct",
        "durable": True,
    }), ((), {
        "exchange": "hive.dead.letter",
        "exchange_type": "direct",
        "durable": True,
    })]
    assert mock.basic_qos.call_log == [((), {
        "prefetch_count": 1,
    })]
    assert mock.queue_declare.call_log == [((
        "x.arr.pirates",
    ), {
        "durable": True,
    }), ((
        "arr.pirates",
    ), {
        "durable": True,
        "arguments": {
            "x-dead-letter-exchange": "hive.dead.letter",
        },
    })]
    assert mock.queue_bind.call_log == [((), {
        "queue": "x.arr.pirates",
        "exchange": "hive.dead.letter",
        "routing_key": "arr.pirates",
    }), ((), {
        "queue": "arr.pirates",
        "exchange": "hive.events",
        "routing_key": "arr.pirates",
    })]

    assert len(mock.basic_consume.call_log) == 1
    assert len(mock.basic_consume.call_log[0]) == 2
    got_callback = mock.basic_consume.call_log[0][1]["on_message_callback"]
    assert mock.basic_consume.call_log == [((), {
        "queue": "arr.pirates",
        "on_message_callback": got_callback,
    })]
    assert on_message_callback.call_log == []
    assert mock.basic_ack.call_log == []

    expect_method = type("method", (), {"delivery_tag": 5})
    expect_body = b'{"hello":"W0RLD"}'
    got_callback(channel._pika, expect_method, expect_properties, expect_body)

    assert len(on_message_callback.call_log) == 1
    assert len(on_message_callback.call_log[0]) == 2
    assert len(on_message_callback.call_log[0][0]) == 2
    message = on_message_callback.call_log[0][0][1]
    assert on_message_callback.call_log == [((channel, message), {})]
    assert message.method is expect_method
    assert message.properties is expect_properties
    assert message.body is expect_body

    assert mock.basic_ack.call_log == [((), {"delivery_tag": 5})]


def test_publish_fanout_event():
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_publish = MockMethod()

    channel = Channel(pika=mock)
    channel.publish_event(
        message={
            "bonjour": "madame",
        },
        routing_key="egg.nog",
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.egg.nog",
        "exchange_type": "fanout",
        "durable": True,
    })]
    assert mock.basic_publish.call_log == [((), {
        "exchange": "hive.egg.nog",
        "routing_key": "",
        "body": b'{"bonjour": "madame"}',
        "properties": expect_properties,
        "mandatory": False,
    })]


@pytest.mark.parametrize("callback_cls", (MockCallbackV1, MockCallbackV2))
@pytest.mark.filterwarnings(MockCallbackV1._warning_filter)
def test_consume_fanout_events(callback_cls):
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_qos = MockMethod()
    mock.queue_declare = MockMethod(
        returns=type("Result", (), dict(
            method=type("Method", (), dict(
                queue="TeStQuEu3")))))
    mock.queue_bind = MockMethod()
    mock.basic_consume = MockMethod()
    on_message_callback = callback_cls()
    mock.basic_ack = MockMethod()

    channel = Channel(pika=mock)
    channel.consume_events(
        queue="arr.pirates",
        on_message_callback=on_message_callback,
    )

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.arr.pirates",
        "exchange_type": "fanout",
        "durable": True,
    }), ((), {
        "exchange": "hive.dead.letter",
        "exchange_type": "direct",
        "durable": True,
    })]
    assert mock.basic_qos.call_log == [((), {
        "prefetch_count": 1,
    })]
    assert mock.queue_declare.call_log == [((
        "x.pytest.arr.pirates",
    ), {
        "durable": True,
    }), ((
        "pytest.arr.pirates",
    ), {
        "exclusive": True,
        "arguments": {
            "x-dead-letter-exchange": "hive.dead.letter",
        },
    })]
    assert mock.queue_bind.call_log == [((), {
        "queue": "x.pytest.arr.pirates",
        "exchange": "hive.dead.letter",
        "routing_key": "pytest.arr.pirates",
    }), ((), {
        "queue": "TeStQuEu3",
        "exchange": "hive.arr.pirates",
    })]

    assert len(mock.basic_consume.call_log) == 1
    assert len(mock.basic_consume.call_log[0]) == 2
    got_callback = mock.basic_consume.call_log[0][1]["on_message_callback"]
    assert mock.basic_consume.call_log == [((), {
        "queue": "TeStQuEu3",
        "on_message_callback": got_callback,
    })]
    assert on_message_callback.call_log == []
    assert mock.basic_ack.call_log == []

    expect_method = type("method", (), {"delivery_tag": 5})
    expect_body = b'{"hello":"W0RLD"}'
    got_callback(channel._pika, expect_method, expect_properties, expect_body)

    assert len(on_message_callback.call_log) == 1
    assert len(on_message_callback.call_log[0]) == 2
    assert len(on_message_callback.call_log[0][0]) == 2
    message = on_message_callback.call_log[0][0][1]
    assert on_message_callback.call_log == [((channel, message), {})]
    assert message.method is expect_method
    assert message.properties is expect_properties
    assert message.body is expect_body

    assert mock.basic_ack.call_log == [((), {"delivery_tag": 5})]


def test_tell_user():
    mock = MockPika()
    mock.exchange_declare = MockMethod()
    mock.basic_publish = MockMethod()

    channel = Channel(pika=mock)
    channel.tell_user("bonjour!")

    assert mock.exchange_declare.call_log == [((), {
        "exchange": "hive.chat.messages",
        "exchange_type": "fanout",
        "durable": True,
    })]

    assert len(mock.basic_publish.call_log) == 1
    assert len(mock.basic_publish.call_log[0]) == 2
    got_body = mock.basic_publish.call_log[0][1]["body"]
    assert mock.basic_publish.call_log == [((), {
        "exchange": "hive.chat.messages",
        "routing_key": "",
        "body": got_body,
        "properties": expect_properties,
        "mandatory": False,
    })]

    assert isinstance(got_body, bytes)
    message = json.loads(got_body)

    uuid = message.pop("uuid")
    assert isinstance(uuid, str)
    uuid_re = "-".join(f"[0-9a-f]{{{n}}}" for n in (8, 4, 4, 4, 12))
    uuid_re = f"^{uuid_re}$"
    assert re.match(uuid_re, uuid)

    timestamp = message.pop("timestamp")
    assert isinstance(timestamp, str)
    timestamp_re = r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{6}\+00:00$"
    assert re.match(timestamp_re, timestamp)

    assert message == {
        "sender": "hive",
        "text": "bonjour!",
    }



import inspect
import logging

from typing import Any

from pika import DeliveryMode

logger = logging.getLogger(__name__)


def publish_may_drop(kwargs: dict[str, Any]):
    """Unroutable messages are silently dropped.
    """
    _ensure_kwarg(kwargs, "mandatory", False)


def publish_must_succeed(kwargs: dict[str, Any]):
    """The message is delivered or an exception is raised.
    """
    _ensure_kwarg(kwargs, "mandatory", True)
    messages_persist_across_broker_restarts(kwargs)


def messages_persist_across_broker_restarts(kwargs: dict[str, Any]):
    """The message won't get lost if the broker is restarted (part 1 of 2).
    """
    _ensure_kwarg(kwargs, "delivery_mode", DeliveryMode.Persistent)


_Unset = object()


def _ensure_kwarg(kwargs: dict[str, Any], name: str, required_value: Any):
    got_value = kwargs.get(name, _Unset)
    if got_value is _Unset:
        kwargs[name] = required_value
        return

    frame = inspect.currentframe()
    while frame and frame.f_code.co_filename == __file__:
        frame = frame.f_back
    func_name = frame.f_code.co_name

    if got_value is not required_value:
        raise ValueError(f"{func_name}:{name}={got_value!r}")

    logger.warning('%s: "%s" keyword argument is deprecated',
                   func_name, name, stacklevel=2)



import json

from dataclasses import dataclass

from pika.spec import Basic, BasicProperties


@dataclass
class Message:
    method: Basic.Deliver
    properties: BasicProperties
    body: bytes

    @property
    def content_type(self) -> str:
        return self.properties.content_type

    def json(self):
        if self.content_type != "application/json":
            raise ValueError(self.content_type)
        return json.loads(self.body)



import inspect
import json
import logging
import os
import sys
import warnings

from functools import cache, cached_property
from typing import Callable, Optional

from pika import BasicProperties, DeliveryMode

from . import semantics
from .message import Message
from .schemas import ChatMessage
from .wrapper import WrappedPikaThing

logger = logging.getLogger(__name__)
d = logger.debug


class Channel(WrappedPikaThing):
    """The primary entry point for interacting with Hive's message bus.
    """
    # QUEUES are declared by their consuming service

    # CONSUME_* methods process to completion or dead-letter the message

    # REQUESTS are things we're asking to be done:
    # - Each request queue has exactly one consuming service
    # - Publish delivers the message or raises an exception
    # - Consume processes to completion or dead-letters the message

    def publish_request(self, **kwargs):
        return self._publish_direct(
            self.requests_exchange,
            **kwargs
        )

    def consume_requests(self, **kwargs):
        return self._consume_direct(
            self.requests_exchange,
            **kwargs
        )

    # EVENTS are things that have happened:
    # - DIRECT EVENTS have the same semantics as requests
    #
    # - FANOUT EVENTS are different:
    #   - Transient events fan-out to zero-many consuming services
    #   - Publish drops messages with no consumers

    def publish_event(self, *, mandatory: bool = False, **kwargs):
        if mandatory:
            return self._publish_direct(
                self.direct_events_exchange,
                **kwargs
            )
        return self._publish_fanout(**kwargs)

    def consume_events(
            self,
            queue: str,
            mandatory: bool = False,
            **kwargs
    ):
        if mandatory:
            return self._consume_direct(
                self.direct_events_exchange,
                queue=queue,
                **kwargs
            )
        return self._consume_fanout(queue, **kwargs)

    # Lower level handlers for REQUESTS and EVENTS

    def _publish_direct(self, exchange: str, **kwargs):
        semantics.publish_must_succeed(kwargs)
        return self._publish(exchange=exchange, **kwargs)

    def _publish_fanout(self, routing_key: str, **kwargs):
        semantics.publish_may_drop(kwargs)
        exchange = self._fanout_exchange_for(routing_key)
        return self._publish(exchange=exchange, **kwargs)

    def _consume_direct(
            self,
            exchange: str,
            *,
            queue: str,
            on_message_callback: Callable,
    ):
        self.queue_declare(
            queue,
            dead_letter_routing_key=queue,
            durable=True,
        )

        self.queue_bind(
            queue=queue,
            exchange=exchange,
            routing_key=queue,
        )

        return self._basic_consume(queue, on_message_callback)

    def _consume_fanout(
            self,
            routing_key: str,
            *,
            on_message_callback: Callable,
    ):
        exchange = self._fanout_exchange_for(routing_key)
        if (queue := self.consumer_name):
            queue = f"{queue}.{routing_key}"

        queue = self.queue_declare(
            queue,
            dead_letter_routing_key=queue or routing_key,
            exclusive=True,
        ).method.queue

        self.queue_bind(
            queue=queue,
            exchange=exchange,
        )

        return self._basic_consume(queue, on_message_callback)

    @cached_property
    def consumer_name(self):
        """Name for per-consumer fanout queues to this channel.
        May be overwritten or overridden (you'll actually have
        to if more than one channel per process consumes the
        same fanout "queue").
        """
        return ".".join(
            part for part in os.path.basename(sys.argv[0]).split("-")
            if part != "hive"
        )

    # Exchanges

    @cache
    def _fanout_exchange_for(self, routing_key: str) -> str:
        return self._hive_exchange(
            exchange=routing_key,
            exchange_type="fanout",
            durable=True,
        )

    @cached_property
    def direct_events_exchange(self) -> str:
        return self._hive_exchange(
            exchange="events",
            exchange_type="direct",
            durable=True,
        )

    @cached_property
    def requests_exchange(self) -> str:
        return self._hive_exchange(
            exchange="requests",
            exchange_type="direct",
            durable=True,
        )

    @cached_property
    def dead_letter_exchange(self) -> str:
        return self._hive_exchange(
            exchange="dead.letter",
            exchange_type="direct",
            durable=True,
        )

    def _hive_exchange(self, exchange: str, **kwargs) -> str:
        name = f"hive.{exchange}"
        self.exchange_declare(exchange=name, **kwargs)
        return name

    # Queues

    def queue_declare(
            self,
            queue: str,
            *,
            dead_letter_routing_key: Optional[str] = None,
            arguments: Optional[dict[str, str]] = None,
            **kwargs
    ):
        if dead_letter_routing_key:
            DLX_ARG = "x-dead-letter-exchange"
            if arguments:
                if DLX_ARG in arguments:
                    raise ValueError(arguments)
                arguments = arguments.copy()
            else:
                arguments = {}

            dead_letter_queue = f"x.{dead_letter_routing_key}"
            self._pika.queue_declare(
                dead_letter_queue,
                durable=True,
            )

            dead_letter_exchange = self.dead_letter_exchange
            self.queue_bind(
                queue=dead_letter_queue,
                exchange=dead_letter_exchange,
                routing_key=dead_letter_routing_key,
            )

            arguments[DLX_ARG] = dead_letter_exchange

        return self._pika.queue_declare(
            queue,
            arguments=arguments,
            **kwargs
        )

    def _publish(
            self,
            *,
            message: bytes | dict,
            exchange: str = "",
            routing_key: str = "",
            content_type: Optional[str] = None,
            delivery_mode: DeliveryMode = DeliveryMode.Persistent,
            mandatory: bool = True,
    ):
        payload, content_type = self._encapsulate(message, content_type)
        return self.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=payload,
            properties=BasicProperties(
                content_type=content_type,
                delivery_mode=delivery_mode,  # Persist across broker restarts.
            ),
            mandatory=mandatory,  # Don't fail silently.
        )

    @staticmethod
    def _encapsulate(
            msg: bytes | dict,
            content_type: Optional[str],
    ) -> tuple[bytes, str]:
        """Prepare messages for transmission.
        """
        if not isinstance(msg, bytes):
            return json.dumps(msg).encode("utf-8"), "application/json"
        if not content_type:
            raise ValueError(f"content_type={content_type}")
        return msg, content_type

    @property
    def prefetch_count(self):
        return getattr(self, "_prefetch_count", None)

    @prefetch_count.setter
    def prefetch_count(self, value):
        if self.prefetch_count == value:
            return
        if self.prefetch_count is not None:
            raise ValueError(value)
        self.basic_qos(prefetch_count=value)
        self._prefetch_count = value

    def _prepare_omcb(self, cb: Callable) -> Callable:
        """Prepare on_message_callback.
        """
        sig = inspect.signature(cb)
        num_params = len(sig.parameters)
        if num_params == 2:
            return cb
        if num_params != 4:
            raise TypeError(cb)

        warnings.warn(DeprecationWarning(
            "Pika-style on-message callbacks are deprecated"
        ))

        def _wrapped_callback(channel: Channel, msg: Message):
            return cb(channel, msg.method, msg.properties, msg.body)

        return _wrapped_callback

    def _basic_consume(
            self,
            queue: str,
            on_message_callback: Callable,
    ):
        on_message_callback = self._prepare_omcb(on_message_callback)

        self.prefetch_count = 1  # Receive one message at a time.

        def _wrapped_callback(channel: Channel, message: Message):
            delivery_tag = message.method.delivery_tag
            try:
                result = on_message_callback(channel, message)
                channel.basic_ack(delivery_tag=delivery_tag)
                return result
            except Exception as e:
                channel.basic_reject(delivery_tag=delivery_tag, requeue=False)
                logged = False
                try:
                    if isinstance(e, NotImplementedError) and e.args:
                        traceback = e.__traceback__
                        while (next_tb := traceback.tb_next):
                            traceback = next_tb
                        code = traceback.tb_frame.f_code
                        try:
                            func = code.co_qualname
                        except AttributeError:
                            func = code.co_name  # Python <=3.10
                        logger.warning("%s:%s:UNHANDLED", func, e)
                        logged = True

                except Exception:
                    logger.exception("NESTED EXCEPTION")
                if not logged:
                    logger.exception("EXCEPTION")

        return self.basic_consume(
            queue=queue,
            on_message_callback=_wrapped_callback,
        )

    def basic_consume(
            self,
            queue: str,
            on_message_callback,
            *args,
            **kwargs
    ):
        def _wrapped_callback(channel, *args, **kwargs):
            assert channel is self._pika
            return on_message_callback(self, Message(*args, **kwargs))

        return self._pika.basic_consume(
            queue=queue,
            on_message_callback=_wrapped_callback,
            *args,
            **kwargs
        )

    # Utilities

    def tell_user(self, *args, **kwargs):
        """Send a message to the user via the chat UI.
        """
        self.publish_event(
            message=ChatMessage(*args, **kwargs).json(),
            routing_key="chat.messages",
        )


class PublisherChannel:
    def __init__(self, invoker, channel):
        self._invoker = invoker
        self._channel = channel

    def __getattr__(self, attr):
        result = getattr(self._channel, attr)
        if not callable(result):
            return result
        return PublisherInvoker(self._invoker, result)


class PublisherInvoker:
    def __init__(self, invoker, func):
        self._invoke = invoker
        self._func = func

    def __call__(self, *args, **kwargs):
        return self._invoke(self._func, *args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass, field, fields
from datetime import datetime, timezone
from typing import Any, Optional
from uuid import RFC_4122, UUID, uuid4


@dataclass
class ChatMessage:
    text: Optional[str] = None
    html: Optional[str] = None
    sender: str = "hive"
    timestamp: str | datetime = field(
        default_factory=lambda: datetime.now(tz=timezone.utc))
    uuid: str | UUID = field(default_factory=uuid4)
    _unhandled: Optional[dict[str, Any]] = None

    def __post_init__(self):
        if not self.text and not self.html:
            raise ValueError

        if not isinstance(self.timestamp, datetime):
            self.timestamp = datetime.fromisoformat(self.timestamp)

        if not isinstance(self.uuid, UUID):
            self.uuid = UUID(self.uuid)

        if self.uuid.variant != RFC_4122:
            raise ValueError(self.uuid)
        if self.uuid.version != 4:
            raise ValueError(self.uuid)

    @classmethod
    def json_keys(cls) -> list[str]:
        names = (field.name for field in fields(cls))
        return [name for name in names if name[0] != "_"]

    @classmethod
    def from_json(cls, message: dict[str, Any]) -> ChatMessage:
        """Ultra-strict from-(deserialized)-JSON constructor.
        """
        if type(message) is not dict:
            raise TypeError
        if any(type(key) is not str for key in message.keys()):
            raise TypeError

        unhandled = message.copy()
        keys = cls.json_keys()
        values = [unhandled.pop(key, "") for key in keys]

        if any(type(value) is not str for value in values):
            raise TypeError

        kwargs = dict(zip(keys, values))
        if unhandled:
            kwargs["_unhandled"] = unhandled
        return cls(**kwargs)

    @property
    def has_unhandled_fields(self):
        return bool(self._unhandled)

    def json(self) -> dict[str, Any]:
        items = ((key, getattr(self, key)) for key in self.json_keys())
        return dict((key, str(value)) for key, value in items if value)
